! graph_partition_mpi.f90 --- generic graph partition method interface
!
! Copyright  (C)  2010  Thomas Jahns <jahns@dkrz.de>
!
! Version: 1.0
! Keywords: graph partitioning
! Author: Thomas Jahns <jahns@dkrz.de>
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
! Commentary:
!
! This is currently only a convenient wrapper of ParMeTis, other
! heuristics are to follow later.
!
! Code:
!
MODULE graph_partition_mpi
  USE iso_c_binding, ONLY: c_int
  USE scales_ppm_base, ONLY: abort_ppm
  INCLUDE 'mpif.h'
#include <ppm.inc>
  EXTERNAL :: parmetis_v3_partkway
  PUBLIC :: graph_partition
CONTAINS
  SUBROUTINE graph_partition(num_vertices, edge_list_lens, &
       edge_lists, partition_out, comm, &
       balance, num_vertex_weights, vertex_weights, edge_weights)
    INTEGER(ppm_idx), INTENT(in) :: num_vertices
    INTEGER(ppm_idx), INTENT(in) :: edge_list_lens(:)
    INTEGER(ppm_idx), INTENT(in) :: edge_lists(:)
    INTEGER, INTENT(out) :: partition_out(*)
    INTEGER, INTENT(in) :: comm
    REAL(4), OPTIONAL, INTENT(in) :: balance(:, :)
    INTEGER, OPTIONAL, INTENT(in) :: num_vertex_weights
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: vertex_weights(:)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: edge_weights(:)
    INTEGER(c_int) :: wgtflag
    INTEGER :: comm_size, comm_rank, ierror, i
    INTEGER, ALLOCATABLE :: vtxdist(:)
    INTEGER :: metis_options(0:2), edge_cut
    INTEGER, PARAMETER :: max_msg_len=1024
    CHARACTER(len=max_msg_len) :: msg

    CALL mpi_comm_size(comm, comm_size, ierror)
    IF (ierror /= MPI_SUCCESS) THEN
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF
    CALL mpi_comm_rank(comm, comm_rank, ierror)
    IF (ierror /= MPI_SUCCESS) THEN
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF

    ! build table of node distribution
    ALLOCATE(vtxdist(0:comm_size))

    i = INT(num_vertices)
    CALL mpi_allgather(i, 1, MPI_INTEGER, &
         vtxdist(1:comm_size), 1, MPI_INTEGER, comm, ierror)
    IF (ierror /= MPI_SUCCESS) then
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF

    vtxdist(0) = 1
    vtxdist(comm_rank + 1) = i
    DO i = 1, comm_size
      vtxdist(i) = vtxdist(i) + vtxdist(i - 1)
    END DO

    wgtflag = 0
    IF (PRESENT(vertex_weights)) wgtflag = 2
    IF (PRESENT(edge_weights)) wgtflag = IOR(wgtflag, 1)

    metis_options(0) = 0
    ! FIXME: handle balance not present case
    ! FIXME: the 4 needs to be replaced with sp -> import mo_kinds
    CALL parmetis_v3_partkway(vtxdist, edge_list_lens, edge_lists, &
         vertex_weights, edge_weights, wgtflag, 1, num_vertex_weights, &
         comm_size, balance, REAL(1.05, 4), metis_options, edge_cut, &
         partition_out, comm)
  END SUBROUTINE graph_partition
END MODULE graph_partition_mpi
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-default: "bsd"
! End:
